/*
 * Component Name: EIF
 *
 * $Date: 1996/12/10 20:50:46 $
 *
 * $Source: 
 *
 * $Revision: 1.8 $
 *
 * Description:
 *      TEC agent communication library for non-TME agents
 *
 * (C) COPYRIGHT TIVOLI Systems, Inc. 1994
 * Unpublished Work
 * All Rights Reserved
 * Licensed Material - Property of TIVOLI Systems, Inc.
 */
#if defined(WIN32) || defined(OS2)
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef WIN32
#       include <direct.h>
#       include <winsock.h>
#else
#  ifdef OS2
#    include <stdlib.h>
#    include <sys/socket.h>
#    include <types.h>
#    include <netinet\in.h>
#    include <netdb.h>
#    include <signal.h>
#    include <io.h>
#    include <direct.h>
#  else
#       include <stdlib.h>
#       include <unistd.h>
#       include <ctype.h>
#  endif
#endif

#include "agent_comm.h"
#include "eif_common.h"
#include "slist.h"

#if defined(WIN32)
#       define read _hread
#       define write _hwrite
#       define close CloseHandle
#       define mkdir _mkdir
#       define fstat _fstat
#       define strdup _strdup
#else
#       if !defined(AIX) && !defined(OS2)
extern char *strdup(const char *);
#       endif
#       if !defined(SVR4) && !defined(OS2)
extern int ftruncate(int fildes, off_t length);
#       endif
#       if defined(OS2)
#define ftruncate _chsize
#define HFILE    int
#define DWORD  int
#define HANDLE  int
#define MAX_PATH _MAX_PATH
#       endif
#endif

extern void fatal_memory_error();

#if defined(__hpux)
int MkDirR(char *Dir);
#else
static int MkDirR(char *Dir);
#endif


#if defined(WIN32)
static int
pc_truncate(char *path, long size)
{
  int fh;
  int result = -1;

  if( (fh = _open( path, _O_RDWR | _O_CREAT, _S_IREAD | _S_IWRITE ))  != -1 ){
	result = _chsize( fh, size );
	_close( fh );
  }
  return result;
}
#endif

static void
trunc_at_top(char* path, int offset)
{
  int n;
  int size = 0;
  char buf[4096];
	
#ifndef WIN32
  int fd_to = open(path, O_RDWR);
  int fd_from = open(path, O_RDWR);
  if (fd_from < 0 || fd_to < 0) {
	return;
  }
#else
  SECURITY_ATTRIBUTES  sa;
  HANDLE fd_from;
  HANDLE fd_to;
  sa.nLength = sizeof(SECURITY_ATTRIBUTES);
  sa.lpSecurityDescriptor = NULL;
  sa.bInheritHandle = TRUE;
  fd_from = CreateFile(path, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_WRITE | FILE_SHARE_READ, &sa,
			           OPEN_ALWAYS,FILE_ATTRIBUTE_NORMAL, NULL);
  fd_to = CreateFile(path, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_WRITE | FILE_SHARE_READ, &sa,
			         OPEN_ALWAYS,FILE_ATTRIBUTE_NORMAL, NULL);
  if (fd_from == INVALID_HANDLE_VALUE || fd_to == INVALID_HANDLE_VALUE) {
	return;
  }
#endif
	
	
#ifndef WIN32
  lseek((HFILE)fd_from, offset, SEEK_SET);
  while ((n = read(fd_from, buf, 4096))) {
	write(fd_to, buf, n);
	size += n;
  }
  ftruncate(fd_to, size);
  close(fd_from);
  close(fd_to);

#else
  _llseek((HFILE)fd_from, offset, FILE_BEGIN);
  while ((n = _hread((HFILE)fd_from, buf, 4096))) {
	_hwrite((HFILE) fd_to, buf, n);
	size += n;
  }
  _llseek((HFILE)fd_to, size, FILE_BEGIN);
  SetEndOfFile(fd_to);
  CloseHandle(fd_from);
  CloseHandle(fd_to);
}
#endif

/* Since close is used so often, just redefine it here to
   avoid a lot of ifdef's elsewhere.
   */
#ifdef WIN32
#define close(x) CloseHandle(x)
#endif

static char *
get_default_buffer_path()
{
	char			*retval;
	char			*env;
	static char	exp[MAXPATHLEN];

#	ifndef WIN32
		retval = "/etc/Tivoli/tec/cache";
#	else
		retval = "cache.dat";
#	endif

	if ((env = tec_agent_getenv("BufEvtPath")) != NULL) {
		retval = env;

#		ifdef WIN32
			if (ExpandEnvironmentStrings(env, exp, MAXPATHLEN) != 0)
				retval = exp;
#		endif
	}
	return(retval);
}

/*
 * This function does the common work for both tec_buffer_event() and
 * tec_buffer_event_to_file().
 */

int 
tec_buffer_event_to_file(char *message, char *BufEvtPath)
{
  char *env = NULL;
  int MsgLen = strlen(message);

#ifndef WIN32
  int BufEvtFD;
  int CpFromFD;
  int CpToFD;
#else
  SECURITY_ATTRIBUTES  sa;
  HANDLE BufEvtFD;
  HANDLE CpFromFD;
  HANDLE CpToFD;
#endif

  int BufEvtMaxSize = 64*1024;  /* max size, in 1K chunks, of buffer file */
  int BufEvtShrinkSize = 8*1024; /* size, in 1K chunks, to shrink buffer on overflow */
  int BufEvtShrinkBlk = 16*1024; /* copy blk size, 1K chunks, to use when shrinking */
  struct stat BufEvtStat;
  char *ShrinkBuf = NULL;       /* Buffer used when shrinking the cache file */
  char *ShrinkBufPtr = NULL;
  off_t EventSeekPos;           /* Offset to begin shrinking at */
  char *ShrinkBufEventStartPtr = NULL;
  int NumBytesToRd;
  int BytesRd;
  int TotBytesRd;
  int BytesWrt;
  int RetVal;
  char *Dir;
  char *DirEnd;
  int DirVal=0;
  char *Separator = "\001";
  int SeparatorLen = strlen(Separator);
  DWORD derror;
#if defined(OS2)
  int oflag = O_RDWR | O_APPEND | O_CREAT;
#else
  int oflag = O_RDWR | O_APPEND | O_CREAT | O_BINARY;
#endif
#if !defined(WIN32) && !defined(OS2)
  int pmode = S_IRWXU | S_IRGRP | S_IROTH;
#else
#  if defined(OS2)
  int pmode = S_IREAD | S_IWRITE;
#  else
  int pmode = _O_RDWR;
  char exp[MAX_PATH];
#  endif
#endif


  /* return if BufferEvents was specified, but is not yes or YES */
  if ( (env = tec_agent_getenv("BufferEvents")) != NULL ) {
	if ( strcmp(env, "yes") != 0 && strcmp(env, "YES") != 0 ) {
	  return(-1);
	}
  }

  /* Event adapter buffer file's max size in K */
  if ( (env = tec_agent_getenv("BufEvtMaxSize")) != NULL ) {
	if (isdigit(*env)) {
	  BufEvtMaxSize = atoi(env) * 1024; /* converted to bytes */
	}
  }

  /* Size, in K, to shrink adapter file's buffer when MaxSize exceeded */
  if ( (env = tec_agent_getenv("BufEvtShrinkSize")) != NULL ) {
	if (isdigit(*env)) {
	  BufEvtShrinkSize = atoi(env) * 1024; /* converted to bytes */
	}
  }

  /* Size, in K, of blocks used to copy data when shrinking adapter's
   * buffer. This has more to do with Memory mgmt then disk mgmt. */
  if ( (env = tec_agent_getenv("BufEvtShrinkBlk")) != NULL ) {
	if (isdigit(*env)) {
	  BufEvtShrinkBlk = atoi(env) * 1024; /* converted to bytes */
	}
  }

  /*
   * Check if event should not be cached because it is filtered.
   * This is different then the 'normal' event filtering and special
   * ONLY for event that are to be cached. It uses a separate filter
   * then the normal event filtering and is meant to be more restrictive.
   * Events may be important enough to send back but not important
   * enough if they are *first cached* !!
   */
  if (tec_filter_event(message, FILTER_EVENT_CACHE_TYPE)) {
		return(0);
  }

	/*
	 * If no buffer path specified, look it up.
	 */
	if (BufEvtPath == NULL)
		BufEvtPath = get_default_buffer_path();

  /* create the file's directory if it doesn't exist */
  Dir = strdup(BufEvtPath);
  if ((DirEnd = strrchr(Dir, '/')) != NULL) {
	*DirEnd = 0;                /* terminate shortened path */
	if ((DirVal=MkDirR(Dir)) != 0) {
	  return(DirVal);
	}
  }
  free(Dir);

#ifndef WIN32
  if ((BufEvtFD=open(BufEvtPath, oflag, pmode)) == -1) {
		return(-1);
  }

  if (fstat(BufEvtFD, &BufEvtStat) == -1) {
	close(BufEvtFD);
	return(-1);
  }

#else
  sa.nLength = sizeof(SECURITY_ATTRIBUTES);
  sa.lpSecurityDescriptor = NULL;
  sa.bInheritHandle = TRUE;
  BufEvtFD = CreateFile(BufEvtPath, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_WRITE | FILE_SHARE_READ, &sa,
			            OPEN_ALWAYS,FILE_ATTRIBUTE_NORMAL, NULL);
  if (BufEvtFD == INVALID_HANDLE_VALUE)
	return(-1);
  if ((BufEvtStat.st_size = GetFileSize(BufEvtFD, NULL)) == 0xFFFFFFFF) {
	CloseHandle(BufEvtFD);
	derror = GetLastError();
	return(-1);
  }
#endif
  /* If the event buffer cache will overflow then shrink it.
   * Deleting data anywhere in a file is not supported by
   * ftruncate(). The data must be copied down so the end
   * of the file can be deleted.
   * 'BufEvtShrinkSize' is an estimate of how much to shrink
   * the file since this may be in the middle of an event.
   * Search for the start of the next event and start shrinking.
   * BufEvtFD is closed and reopened after the file is shrunk
   * as a precaution since I'm not sure if all OS's handle
   * append mode on a file which other FD's are modifying.
   */
  if ((BufEvtStat.st_size + MsgLen) > BufEvtMaxSize) {
	close(BufEvtFD);

	if (BufEvtShrinkSize > BufEvtStat.st_size) {
	  /* shrinking the whole file so just truncate it */
#ifndef WIN32
		/*************************************************************/
		/* Just open & truncate the file instead of using O_TRUNC    */
		/* option on the OPEN. ftruncate discards the file contents. */
		/*************************************************************/
		if ((BufEvtFD=open(BufEvtPath, O_RDWR, 0)) == -1) {
		return(-1);
		}
		if (ftruncate(BufEvtFD,0) == -1) /* truncate the file */
		{
		close(BufEvtFD);                                  
		return(-1);                                   
		}
#else
	  BufEvtFD = CreateFile(BufEvtPath, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_WRITE | FILE_SHARE_READ, &sa,
			                CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL, NULL);
	  if (BufEvtFD == INVALID_HANDLE_VALUE)
		return(-1);
#endif  
	} 
	else {
	  /* 1024 is if 'BufEvtShrinkSize' is not on an event boundary */
	  if ((ShrinkBuf=(char*)malloc(BufEvtShrinkSize+1024)) == NULL) {
		fatal_memory_error();
	  }
#ifndef WIN32
	  if ((CpToFD=open(BufEvtPath, O_RDWR, 0)) == -1) {
		return(-1);
	  }
	  if ((CpFromFD=open(BufEvtPath, O_RDWR, 0)) == -1) {
		close(CpToFD);
		return(-1);
	  }
#else
	  sa.nLength = sizeof(SECURITY_ATTRIBUTES);
	  sa.lpSecurityDescriptor = NULL;
	  sa.bInheritHandle = TRUE;
	  CpToFD = CreateFile(BufEvtPath, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_WRITE | FILE_SHARE_READ, &sa,
			              OPEN_ALWAYS,FILE_ATTRIBUTE_NORMAL, NULL);
	  if (CpToFD == INVALID_HANDLE_VALUE)
		return(-1);
	  sa.nLength = sizeof(SECURITY_ATTRIBUTES);
	  sa.lpSecurityDescriptor = NULL;
	  sa.bInheritHandle = TRUE;
	  CpFromFD = CreateFile(BufEvtPath, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_WRITE | FILE_SHARE_READ, &sa,
			                OPEN_ALWAYS,FILE_ATTRIBUTE_NORMAL, NULL);
	  if (CpFromFD == INVALID_HANDLE_VALUE)
		return(-1);
		
#endif      
	  /* Establish the place to begin shrinking by locating the
	   * start of the next event.
	   *   read() may return fewer bytes then requested so loop to
	   *   read all requested bytes. */
	  NumBytesToRd = BufEvtShrinkSize+1024;
	  ShrinkBufPtr = ShrinkBuf;
	  TotBytesRd = 0;
	  while (((BytesRd=read((HFILE) CpFromFD, ShrinkBufPtr, NumBytesToRd)) > 0) &&
			 (BytesRd != NumBytesToRd)) {
		TotBytesRd += BytesRd;
		NumBytesToRd -= BytesRd;
		ShrinkBufPtr += BytesRd;
	  }
	  if (BytesRd < 0) {
#ifndef WIN32
		close(CpFromFD);
		close(CpToFD);
#else
		CloseHandle(CpFromFD);
		CloseHandle(CpToFD);
#endif
		return(-1);
	  } else {
		TotBytesRd += BytesRd;
		if (TotBytesRd <= (BufEvtShrinkSize)) {
		  close(CpFromFD);      /* failed to rd enough bytes */
		  close(CpToFD);        /* to collapse the file so   */
		  return(-2);           /* return an error.          */
		}
	  }
	  ShrinkBufEventStartPtr = strtok(ShrinkBuf+BufEvtShrinkSize, Separator);
	  ShrinkBufEventStartPtr += (strlen(ShrinkBufEventStartPtr) + SeparatorLen);
	  EventSeekPos = ShrinkBufEventStartPtr - ShrinkBuf;
#ifndef WIN32
	  if (lseek(CpFromFD, EventSeekPos, SEEK_SET) == -1) 
#else
	  if (_llseek((HFILE) CpFromFD, EventSeekPos, FILE_BEGIN) == -1) 
#endif
	  {
	close(CpFromFD);
	close(CpToFD);
	return(-1);
	  }
	  TotBytesRd = 0;
	  ShrinkBufPtr = ShrinkBuf;
	  while ((BytesRd=read((HFILE) CpFromFD, ShrinkBufPtr, BufEvtShrinkSize)) > 0) {
	TotBytesRd += BytesRd;
	if ((BytesWrt=write((HFILE) CpToFD, ShrinkBufPtr, BytesRd)) != BytesRd) {
	  close(CpFromFD);
	  close(CpToFD);
	  return(-1);
	}
	  }
#ifndef WIN32
	  if ((RetVal=ftruncate((HFILE)CpToFD, TotBytesRd)) != 0) {
	close(CpFromFD);
	close(CpToFD);
	return(-1);        
	  }
	  close(CpFromFD);
	  close(CpToFD);
#else
	  close(CpFromFD);
	  close(CpToFD);
	  if ((RetVal=pc_truncate(BufEvtPath, TotBytesRd)) != 0) {
	return(-1);
	  }
#endif
	  free(ShrinkBuf);
	}

#ifndef WIN32
	if ((BufEvtFD=open(BufEvtPath, oflag, pmode)) == -1) {
	  return(-1);
	}
#else
	sa.nLength = sizeof(SECURITY_ATTRIBUTES);
	sa.lpSecurityDescriptor = NULL;
	sa.bInheritHandle = TRUE;
	BufEvtFD = CreateFile(BufEvtPath, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_WRITE | FILE_SHARE_READ, &sa,
			              OPEN_ALWAYS,FILE_ATTRIBUTE_NORMAL, NULL);
	if (BufEvtFD == INVALID_HANDLE_VALUE) {
	  derror = GetLastError();
	  return(-1);
	} 
#endif
  }

#ifdef WIN32
  _llseek((HFILE)BufEvtFD, 0, FILE_END);
#endif
  write((HFILE)BufEvtFD, message, MsgLen); /* cache event to file */
  close(BufEvtFD);

  return(MsgLen); 
} /* do_buffering() */

int
tec_buffer_event(char *message)
{
	return(tec_buffer_event_to_file(message, NULL));
}

void
tec_free_func(void *ptr)
{
  if ( ptr )
	free(ptr);
}

int in_flush_events = 0;

/*
 * The first two input parameters are self-explanatory, while the third one
 * represents a pointer to the appropriate send function which will be used
 * to put events to the TEC server.
 *
 * Flush all events in the event buffer cache (on disk).
 * Try to send each event. If it won't go, then it will
 * will get re-buffered again by tec_buffer_event().
 *
 * As data blocks are read a partial event may be read in.
 * To detect this each potential _next_ event is checked
 * for the 'Separator' string. If the 'Separator' is not
 * found then it may contain a partial event. If the next
 * 'event' is NULL then the read ended on an event boundary
 * otherwise it is a partial event so move it to the front
 * of the buffer and have the read() fill in after it.
 *
 */

int
tec_flush_events_from_file(tec_handle_t handle, char *BufEvtPath)
{
  char *env = NULL;
  tec_handle_common_t     *th = (tec_handle_common_t *) handle;
  char *event;
  char *EventCopy = NULL;
  int EventCopyLen;
  char *NextEvtPtr=NULL;
  int rate = 0;
  int nap=0;
  int sent=0;
  int EvtCnt = 0;
#ifndef WIN32
  int BufEvtFD;
#  ifndef OS2
  struct flock buflock;		/* used to lock cache file */
#   endif
#else
  DWORD  derror;
  HANDLE BufEvtFD;
  int buf_size_left;	/* keep track of buffer left to flush */
#endif
  struct stat BufEvtStat;
  int BufEvtRdBlkLen=65536;   /* default to reading in 64K chunks */
  char *BufEvtBlk=NULL;
  char *BufEvtBlkPtr=NULL;
  char *PartEvtBuf=NULL;
  int RdLen;
  int PartEvtLen;
  int BytesRd;
  int TotBytesRd = 0;
  SECURITY_ATTRIBUTES  sa;

  char *Separator = "\001";
  int SeparatorLen = strlen(Separator);
  char *SeparatorFnd = NULL;



  /* If BufferFlushRate is not specified, then we will use 0
   * which means don't wait at all */
  if ( (env = tec_agent_getenv("BufferFlushRate")) != NULL ) {
	if ( isdigit(*env) )
	  rate = atoi(env);
  }

  /* get the size of the buffer file blocks to read.
   * this MUST be large enough for at least one event. */
  if ( (env = tec_agent_getenv("BufEvtRdBlkLen")) != NULL ) {
	if (isdigit(*env)) {
	  BufEvtRdBlkLen = atoi(env);
	}
  }

	/*
	 * If no buffer path specified, look it up.
	 */
	if (BufEvtPath == NULL)
		BufEvtPath = get_default_buffer_path();

#ifndef WIN32
  if ((BufEvtFD=open(BufEvtPath, O_RDWR, 0)) == -1) {
	return(errno);
  }
#else
  sa.nLength = sizeof(SECURITY_ATTRIBUTES);
  sa.lpSecurityDescriptor = NULL;
  sa.bInheritHandle = TRUE;

  BufEvtFD = CreateFile(BufEvtPath, GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_WRITE | FILE_SHARE_READ, &sa, OPEN_ALWAYS,
			FILE_ATTRIBUTE_NORMAL, NULL);

  if (BufEvtFD == INVALID_HANDLE_VALUE)
	return errno;
#endif
	
#ifndef WIN32
  if (fstat(BufEvtFD, &BufEvtStat) == -1) {
	close(BufEvtFD);
	return(errno);
  }

#  ifndef OS2
  buflock.l_type = F_WRLCK;
  buflock.l_start = 0;
  buflock.l_whence = SEEK_SET;
  buflock.l_len = 0;

  /* Lock the cache file so no one else tries to flush it at the same time 
     If we can't lock it means someone else is flushing, so exit           */
  if (fcntl(BufEvtFD, F_SETLK, &buflock) < 0) {
    close(BufEvtFD);
    return(sent);
  }
#  endif
#else
  if ((BufEvtStat.st_size = GetFileSize(BufEvtFD, NULL)) == 0xFFFFFFFF) {
	CloseHandle(BufEvtFD);
	derror = GetLastError();
	return(derror);
  }

  buf_size_left = BufEvtStat.st_size;

  /* Lock the cache file so no one else tries to flush it at the same time
     If we can't lock it means someone else is flushing, so exit           */
  if (!LockFile(BufEvtFD, 0, 0, buf_size_left, 0)) {
    CloseHandle(BufEvtFD);
    return(sent);
  }
#endif

  if ((BufEvtBlk=(char*) malloc(BufEvtRdBlkLen+1)) == NULL) {
	fatal_memory_error();
  }

  /* only now we can set in_flush_events, and do not return from this
   function without clearing it.
  */
	
  th->in_flush_events = 1;
  BufEvtBlkPtr = BufEvtBlk;
  RdLen = BufEvtRdBlkLen;

  /*
   * During a flush, the event server may go down and this would cause
   * the event to reenter the cache. Thus, we read the cache file only
   * to the size before flushing (as in BufEvtStat) and later shrink
   * the file from the beginning to this size. Anything beyond this size
   * must be due to events reentering the cache (or brand new events
   * if we were threaded).
   */
	
  while ((TotBytesRd < BufEvtStat.st_size)
	 && (BytesRd=read((HFILE) BufEvtFD, BufEvtBlkPtr, RdLen)) > 0)
	{
	  TotBytesRd += BytesRd;

	  /* add terminator to make a real string */      
	  if (TotBytesRd <= BufEvtStat.st_size) {
	BufEvtBlkPtr[BytesRd] = 0;
	  }
	  else {
	/* we have read past the original cache size */
	BufEvtBlkPtr[BytesRd - (TotBytesRd - BufEvtStat.st_size)] = 0;     
	  }
		
	  SeparatorFnd = strstr(BufEvtBlk, Separator);
	  event = strtok(BufEvtBlk, Separator); /* First event */
	  NextEvtPtr = event + strlen(event) + SeparatorLen;

	  while (SeparatorFnd != NULL) {
	EvtCnt++;
	/* Copy the event putting the 'Separator' back. Putting the
	 * Separator back into 'event' is inadequate because that
	 * does not form an event *string* (the string continues to
	 * the end of the read buffer).   */
	EventCopyLen = strlen(event) + SeparatorLen + 1;

	if ((EventCopy=(char*)malloc(EventCopyLen)) == NULL)
	  fatal_memory_error();

	strcpy(EventCopy, event);
	strcat(EventCopy, Separator);

	if (tec_put_event(handle, EventCopy) > 0 )
	  sent++;

#ifdef WIN32
        /* must unlock for NT so trunc_at_top() can open it */
	UnlockFile(BufEvtFD, 0, 0, buf_size_left, 0);
#endif
		
	trunc_at_top(BufEvtPath, strlen(event) + SeparatorLen);

	/* lock it back up - Unix unlocks during trunc_at_top() */
#ifndef WIN32
#  ifndef OS2
	if (fcntl(BufEvtFD, F_SETLK, &buflock) < 0) {
		close(BufEvtFD);
		return(sent);
	}
#  endif
#else
	buf_size_left = buf_size_left - (strlen(event) + SeparatorLen);

	if (!LockFile(BufEvtFD, 0, 0, buf_size_left, 0)) {
		CloseHandle(BufEvtFD);
		return(sent);
	}
#endif

	free(EventCopy);

	SeparatorFnd = strstr(NextEvtPtr, Separator);
	event = strtok(NextEvtPtr, Separator); /* next event */

	if (event != NULL)
	  NextEvtPtr = event + strlen(event) + SeparatorLen;

	/* If rate is defined, and we have sent 'rate' events,
	 * sleep for a minute.  This defines rate events/minute
	 * even though they are sent in bursts
	 */
	if (rate)
	  nap = (EvtCnt % rate);

	if (rate != 0  && nap == 0)
#ifndef WIN32
	  sleep(60);
#else
	Sleep(60*1000);
#endif

	  }

	  /* If a partial event is left in the memory buffer then move it
	   * to the front of the buffer and have the read() fill in after
	   * it.     */
	  if (event != NULL) {
	/* Create a temp buffer to hold the partial event
	 * before copying back into the real buffer just
	 * in case a small buffer was read/specified and
	 * the memcpy() might overlap.
	 * Init the BufEvtBlkPtr and RdLen correctly.
	 */
	PartEvtLen = BufEvtBlkPtr + BytesRd - event;

	if ((PartEvtBuf=(char*)malloc(PartEvtLen)) == NULL)
	  fatal_memory_error();

	memcpy(PartEvtBuf, event, PartEvtLen);
	memcpy(BufEvtBlk, PartEvtBuf, PartEvtLen);
	BufEvtBlkPtr = BufEvtBlk + PartEvtLen;
	RdLen = BufEvtRdBlkLen - PartEvtLen;
	free(PartEvtBuf);
	  }
	  else {
	/* reset the buffer ptr and read length */
	BufEvtBlkPtr = BufEvtBlk;
	RdLen = BufEvtRdBlkLen;
	  }
	}  // end of while

  free(BufEvtBlk);

#ifndef WIN32
#  ifndef OS2
  buflock.l_type = F_UNLCK;
  fcntl(BufEvtFD, F_SETLKW, &buflock);
#  endif
  close(BufEvtFD);
#else
  UnlockFile(BufEvtFD, 0, 0, buf_size_left, 0);
  CloseHandle(BufEvtFD);
#endif
	
  th->in_flush_events = 0;
  return(sent);
}


/*
 * tec_flush_events -
 *
 * Description:
 * Input:
 * handle - the event server handle to send the cached
 *           events to.
 */
int
tec_flush_events(tec_handle_t handle)
{
   return(tec_flush_events_from_file(handle, NULL));

}

/*
 * MkDirR
 *
 * Description:
 *      Recursively make the given directory.
 *      If the mkdir() works or the directory exists then return.
 *      If ENOENT is returned by mkdir(), something in the path
 *      doesn't exist, then remove the last node from the path and
 *      call MkDirR() to try and make that directory.
 *
 * Input:
 *      Dir - the directory path to create.
 */
int MkDirR(char *Dir) {
  char *DirDup;
  char *PathEnd;
  int  RetVal = 0;

#if !defined(WIN32) && !defined(OS2)
  if (mkdir(Dir, (S_IRWXU|S_IRGRP|S_IROTH)) != 0) 
#else
  if (mkdir(Dir) != 0) 
#endif
  {
	switch (errno) {
	case ENOENT:
	  /* part of path doesn't exist so create parent paths
	   * then create _this_ Dir. */
	  DirDup = strdup(Dir);
	  if ((PathEnd = strrchr(DirDup, '/')) != NULL) {
	*PathEnd = 0; /* terminate shortened path */
	if (MkDirR(DirDup) == 0) {
#if !defined(WIN32) && !defined(OS2)
	  if (mkdir(Dir, (S_IRWXU|S_IRGRP|S_IROTH)) != 0)
#else
		  if (mkdir(Dir) != 0)
#endif
		{
		  RetVal = -4;
		}
	} else {
	  RetVal = -2;
	}
	  } else {
	RetVal = -3; /* mkdir err creating first elem of Dir */
	  }
	  free(DirDup);
	  break;
	case EEXIST:
	  RetVal = 0;
	  break;
	default:
	  RetVal = -1;  /* must be some sort of error */
	  break;
	}
  }

  return RetVal;
}

#endif
